@@ -83,6 +83,10 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             return new PopupMenu.PopupMenuItem(opts.label);
         };
 
+        // TODO: When no windows exist on the active workspace, but do on another,
+        // we should detect those cases and offer workspace options if they are pinned. Otherwise,
+        // user needs to switch workspaces just to switch the window back. This should behave this way
+        // when showAllWorkspaces is disabled as well since its a UX problem.
         if (hasWindows) {
             // Monitors
             if (Main.layoutManager.monitors.length > 1) {
@@ -108,6 +112,9 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                     item = createMenuItem({label: _('Only on this workspace')});
                     this.signals.connect(item, 'activate', () => {
                         this.groupState.lastFocused.unstick();
+                        // Always index windows from all workspaces while showAllWorkspaces is enabled
+                        if (this.state.settings.showAllWorkspaces) return;
+                        this.state.removingWindowFromWorkspaces = true;
                         this.state.trigger('removeWindowFromOtherWorkspaces', this.groupState.lastFocused);
                     });
                     this.addMenuItem(item);
@@ -128,17 +135,18 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                         });
                     };
                     for (let i = 0; i < length; i++) {
-                        // Make the index a local letiable to pass to function
+                        // Make the index a local variable to pass to function
                         let j = i;
                         let name = Main.workspace_names[i] ? Main.workspace_names[i] : Main._makeDefaultWorkspaceName(i);
-                        let ws = createMenuItem({label: _(name)});
+                        let menuItem = createMenuItem({label: _(name)});
+                        let ws = this.groupState.lastFocused.get_workspace();
 
-                        if (i === this.state.currentWs) {
-                            ws.setSensitive(false);
+                        if (ws && i === ws.index()) {
+                            menuItem.setSensitive(false);
                         }
 
-                        connectWorkspaceEvent(ws, j);
-                        item.menu.addMenuItem(ws);
+                        connectWorkspaceEvent(menuItem, j);
+                        item.menu.addMenuItem(menuItem);
                     }
                 }
             }
@@ -157,10 +165,9 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         this.signals.connect(item, 'activate', () => this.state.trigger('configureApplet'));
         subMenu.menu.addMenuItem(item);
 
-        item = createMenuItem({label: _('Remove') + " 'Grouped window list'", icon: 'edit-delete'});
-        this.signals.connect(item, 'activate', () => {
-            AppletManager._removeAppletFromPanel(this.state.uuid, this.state.instance_id);
-        });
+        item = createMenuItem({label: _("Remove '%s'").format(_("Grouped window list")), icon: 'edit-delete'});
+        this.signals.connect(item, 'activate', (actor, event) => this.state.trigger('removeApplet', event));
+
         subMenu.menu.addMenuItem(item);
         this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
 
@@ -219,6 +226,15 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             }
         }
 
+        if (Main.gpu_offload_supported && !hasWindows) {
+            item = createMenuItem({label: _("Run with NVIDIA GPU"), icon: 'cpu'});
+
+            this.signals.connect(item, 'activate', () => this.groupState.trigger('launchNewInstance', true));
+
+            this.addMenuItem(item);
+            this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
+        }
+
         // Actions
         tryFn(() => {
             if (!this.groupState.appInfo) return;
@@ -274,19 +290,17 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         // Pin/unpin, shortcut handling
         if (!isWindowBacked) {
-            if (this.state.settings.showPinned !== FavType.none) {
-                let label, icon;
-                if (this.groupState.isFavoriteApp) {
-                    label = _('Unpin from Panel');
-                    icon = 'unpin';
-                } else {
-                    label = _('Pin to Panel');
-                    icon = 'pin';
-                }
-                this.pinToggleItem = createMenuItem({label, icon});
-                this.signals.connect(this.pinToggleItem, 'activate', (...args) => this.toggleFavorite(...args));
-                this.addMenuItem(this.pinToggleItem);
+            let label, icon;
+            if (this.groupState.isFavoriteApp) {
+                label = _('Unpin from Panel');
+                icon = 'unpin';
+            } else {
+                label = _('Pin to Panel');
+                icon = 'pin';
             }
+            this.pinToggleItem = createMenuItem({label, icon});
+            this.signals.connect(this.pinToggleItem, 'activate', (...args) => this.toggleFavorite(...args));
+            this.addMenuItem(this.pinToggleItem);
             if (this.state.settings.autoStart) {
                 let label = this.groupState.autoStartIndex !== -1 ? _('Remove from Autostart') : _('Add to Autostart');
                 item = createMenuItem({label: label, icon: 'insert-object'});
@@ -305,7 +319,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         if (hasWindows) {
             let metaWindowActor = this.groupState.lastFocused.get_compositor_private();
             // Miscellaneous
-            if (metaWindowActor.opacity !== 255) {
+            if (metaWindowActor && metaWindowActor.opacity !== 255) {
                 item = createMenuItem({label: _('Restore to full opacity')});
                 this.signals.connect(item, 'activate', () => metaWindowActor.set_opacity(255));
                 this.addMenuItem(item);
@@ -351,6 +365,9 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                 });
                 this.addMenuItem(item);
                 // Close all
+                // TODO: We should detect if windows from this group are on another workspace
+                // and close windows across all workspaces while showAllWorkspaces is enabled.
+                // Ditto for 'Close others'.
                 item = createMenuItem({label: _('Close all'), icon: 'application-exit'});
                 this.signals.connect(item, 'activate', () => {
                     if (!this.groupState.isFavoriteApp) {
@@ -460,7 +477,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
     destroy() {
         this.signals.disconnectAllSignals();
-        Applet.AppletPopupMenu.prototype.destroy.call(this);
+        super.destroy();
         unref(this, RESERVE_KEYS);
     }
 }
@@ -488,6 +505,21 @@ class HoverMenuController extends PopupMenu.PopupMenuManager {
         this.groupState.disconnect(this.connectId);
         super.destroy();
     }
+
+    _onMenuOpenState(menu, open) {
+        if (open) {
+            if (this._activeMenu && this._activeMenu.isChildMenu(menu)) {
+                this._menuStack.push(this._activeMenu);
+            }
+            this._activeMenu = menu;
+        } else {
+            if (this._menuStack.length > 0) {
+                this._activeMenu = this._menuStack.pop();
+                if (menu.sourceActor)
+                    this._didPop = true;
+            }
+        }
+    }
 }
 
 class WindowThumbnail {
@@ -582,7 +614,8 @@ class WindowThumbnail {
         this.signals.connect(this.button, 'button-release-event', (...args) => this.onCloseButtonRelease(...args));
         this.signals.connect(this.actor, 'button-release-event', (...args) => this.connectToWindow(...args));
 
-        setTimeout(() => this.handleFavorite(), 0);
+
+        this.handleFavorite();
         // Update focused style
         this.onFocusWindowChange();
     }
@@ -613,6 +646,7 @@ class WindowThumbnail {
 
     onLeave() {
         this.entered = false;
+        this.stopClick = false;
         this.actor.remove_style_pseudo_class('selected');
         this.onFocusWindowChange();
         this.button.set_opacity(0);
@@ -658,6 +692,8 @@ class WindowThumbnail {
         this.metaWindow.delete(global.get_current_time());
         if (!this.groupState.metaWindows || this.groupState.metaWindows.length <= 1) {
             this.groupState.trigger('hoverMenuClose');
+        } else {
+            this.groupState.trigger('checkShouldClose');
         }
     }
 
@@ -686,33 +722,29 @@ class WindowThumbnail {
 
     getThumbnail(thumbnailWidth, thumbnailHeight) {
         if (this.groupState.verticalThumbs || !this.state.settings.showThumbs) {
-            this.thumbnailActor.height = 0;
+            this.thumbnailActor.hide();
             return null;
+        } else if (this.thumbnailActor.realized) {
+            this.thumbnailActor.show();
         }
         // Create our own thumbnail if it doesn't exist
-        let isUpdate = false;
         if (this.metaWindowActor) {
-            isUpdate = true;
             this.signals.disconnect('size-changed', this.metaWindowActor);
+        } else {
+            this.metaWindowActor = this.metaWindow.get_compositor_private();
         }
-        this.metaWindowActor = this.metaWindow.get_compositor_private();
-        if (this.metaWindowActor) {
-            let windowTexture = this.metaWindowActor.get_texture();
-            let [width, height] = windowTexture.get_size();
+        if (this.metaWindowActor && !this.metaWindowActor.is_finalized()) {
             this.signals.connect(this.metaWindowActor, 'size-changed', () => this.refreshThumbnail());
 
-            if (this.groupState.isFavoriteApp) {
-                this.signals.connect(this.metaWindowActor, 'destroy', () => {
-                    if (this.willUnmount || !this.groupState.trigger) return;
-                    this.groupState.trigger('removeThumbnailFromMenu', this.metaWindow);
-                    this.metaWindowActor = null;
-                });
-            }
-
+            let windowTexture = this.metaWindowActor.get_texture();
+            if (!windowTexture) return;
+            let [width, height] = windowTexture.get_size();
             let scale = Math.min(1.0, thumbnailWidth / width, thumbnailHeight / height) * global.ui_scale;
             width = Math.round(width * scale);
             height = Math.round(height * scale);
-            if (isUpdate) {
+            if (this.thumbnailActor.child) {
+                this.thumbnailActor.height = height;
+                this.thumbnailActor.width = width;
                 this.thumbnailActor.child.source = windowTexture;
                 this.thumbnailActor.child.width = width;
                 this.thumbnailActor.child.height = height;
@@ -802,7 +834,10 @@ class WindowThumbnail {
     }
 
     hoverPeek(opacity) {
-        if (!this.state.settings.enablePeek || this.state.overlayPreview || this.state.scrollActive) {
+        if (!this.state.settings.enablePeek
+            || this.state.overlayPreview
+            || this.state.scrollActive
+            || (this.metaWindowActor && this.metaWindowActor.is_finalized())) {
             return;
         }
         if (!this.metaWindowActor) {
@@ -861,6 +896,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         super._init.call(this, groupState.trigger('getActor'), state.orientation, 0.5);
         this.state = state;
         this.groupState = groupState;
+        this.setCustomStyleClass("grouped-window-list-thumbnail-menu");
 
         this.connectId = this.groupState.connect({
             hoverMenuClose: () => {
@@ -868,7 +904,28 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
                 this.groupState.set({thumbnailMenuEntered: false});
                 this.close();
             },
-            addThumbnailToMenu: (win) => this.addThumbnail(win),
+            checkShouldClose: () => {
+                // This is called after a close button is clicked. When the menu size changes, it can leave the cursor
+                // outside the menu bounds, and no leave event will be able to correct this situation where the menu is
+                // dangling open and only closable upon hovering over it again. This checks if the cursor is hovering
+                // over the menu and closes it if not.
+                setTimeout(() => {
+                    let [x, y, mask] = global.get_pointer();
+                    let draggedOverActor = global.stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
+                    let parent = draggedOverActor.get_parent();
+                    if (!(parent instanceof St.Widget)) {
+                        this.close(true);
+                    }
+                }, 500);
+            },
+            addThumbnailToMenu: (win) => {
+                if (this.isOpen) {
+                    this.addThumbnail(win);
+                    return;
+                }
+
+                this.queuedWindows.push(win);
+            },
             removeThumbnailFromMenu: (win) => {
                 let index = findIndex(this.appThumbnails, (item) => item.metaWindow === win);
                 if (index > -1) {
@@ -876,16 +933,50 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
                     this.appThumbnails[index] = undefined;
                     this.appThumbnails.splice(index, 1);
                 }
+                index = this.queuedWindows.indexOf(win);
+                if (index > -1) this.queuedWindows.splice(index, 1);
             },
             verticalThumbs: () => {
                 // Preserve the menu's open state after refreshing
                 let {isOpen} = this;
-                this.setVerticalSetting()
+                this.setVerticalSetting();
                 if (isOpen) this.open(true);
             },
+            fileDrag: ({fileDrag}) => {
+                if (fileDrag) {
+                    // When a drag operation from another app is started, no events fire, so we have to grab the
+                    // cursor, find the actor by coordinates, and then look up the thumbnail actor. Do this on a
+                    // 50ms loop until the menu closes so we continue getting data in the absence of events.
+                    this.interval = setInterval(() => {
+                        let [x, y, mask] = global.get_pointer();
+                        let draggedOverActor = global.stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
+                        if (draggedOverActor instanceof Meta.ShapedTexture) {
+                            this.groupState.set({fileDrag: false});
+                            this.close(true);
+                            return;
+                        }
+                        each(this.appThumbnails, function(thumbnail) {
+                            if (thumbnail.thumbnailActor === draggedOverActor) {
+                                Main.activateWindow(thumbnail.metaWindow, global.get_current_time());
+                                return false;
+                            }
+                        });
+                    }, 50);
+                } else if (this.interval) {
+                    clearInterval(this.interval);
+                }
+            }
         });
 
         this.appThumbnails = [];
+        this.queuedWindows = [];
+        this.fullyRefreshThumbnails();
+    }
+
+    addQueuedThumbnails() {
+        if (this.queuedWindows.length === 0) return;
+        each(this.queuedWindows, (win) => this.addThumbnail(win));
+        this.queuedWindows = [];
     }
 
     onButtonPress() {
@@ -912,6 +1003,8 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             timeout = this.state.settings.thumbTimeout;
         }
 
+        this.addQueuedThumbnails();
+
         if (actor != null) {
             this.groupState.set({thumbnailMenuEntered: this.isOpen});
         }
@@ -953,8 +1046,9 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             this.groupState.tooltip.set_text(this.groupState.appName);
             this.groupState.tooltip.show();
         } else {
+            if (force || this.state.settings.onClickThumbs) this.addQueuedThumbnails();
             this.state.set({thumbnailMenuOpen: true});
-            PopupMenu.PopupMenu.prototype.open.call(this, this.state.settings.animateThumbs);
+            super.open(this.state.settings.animateThumbs);
         }
     }
 
@@ -965,17 +1059,22 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             || !this.groupState.tooltip) {
             return;
         }
-        if (!this.groupState.metaWindows || this.groupState.metaWindows.length === 0) {
+        if ((!this.groupState.metaWindows || this.groupState.metaWindows.length === 0)
+            && !this.groupState.tooltip._tooltip.is_finalized()) {
             this.groupState.tooltip.set_text('');
             this.groupState.tooltip.hide();
         }
         if (this.isOpen) {
             this.state.set({thumbnailMenuOpen: false});
-            PopupMenu.PopupMenu.prototype.close.call(this, this.state.settings.animateThumbs);
+            if (!this.actor.is_finalized()) super.close(this.state.settings.animateThumbs);
         }
         for (let i = 0; i < this.appThumbnails.length; i++) {
             this.appThumbnails[i].destroyOverlayPreview();
         }
+
+        if (this.groupState.fileDrag) {
+            this.groupState.set({fileDrag: false});
+        }
     }
 
     onKeyPress(actor, e) {
@@ -1052,7 +1151,6 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             this.destroyThumbnails();
         }
         this.addWindowThumbnails(this.groupState.metaWindows);
-        this.setStyleOptions(false);
     }
 
     destroyThumbnails() {
@@ -1114,37 +1212,20 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         }
     }
 
-    setStyleOptions(skipThumbnailIconResize) {
-        if (this.willUnmount || !this.box) return;
-
-        // The styling cannot be set correctly unless the menu is closed. Fortunately this
-        // can be closed and reopened too quickly for the user to notice.
-        if (this.isOpen) this.close(true);
-
-        this.box.show();
-        this.box.style = null;
-
-        let thumbnailTheme = this.box.peek_theme_node();
-        let padding = thumbnailTheme ? thumbnailTheme.get_horizontal_padding() : null;
-        let thumbnailPadding = padding && (padding > 1 && padding < 21) ? padding : 10;
-        this.box.style = `padding: ${thumbnailPadding / 2}px`;
-        let boxTheme = this.box.peek_theme_node();
-        padding = boxTheme ? boxTheme.get_vertical_padding() : null;
-        let boxPadding = padding && padding > 0 ? padding : 3;
-        this.box.style = `padding: ${boxPadding}px;`;
-
-        if (skipThumbnailIconResize) return;
-
-        if (this.isOpen) this.open();
-    }
-
     setVerticalSetting() {
         if (this.state.orientation === St.Side.TOP || this.state.orientation === St.Side.BOTTOM) {
             this.box.vertical = this.groupState.verticalThumbs || this.state.settings.verticalThumbs;
         } else {
             this.box.vertical = true;
         }
-        this.fullyRefreshThumbnails();
+
+        // Do a full refresh if thumbnails don't exist - this happens when the thumbnail menu
+        // initializes vertically from lack of calculated space, or thumbnails are disabled.
+        if (!this.appThumbnails[0] || !this.appThumbnails[0].thumbnailActor.child) {
+            this.fullyRefreshThumbnails();
+        } else {
+            this.updateThumbnailSize();
+        }
     }
 
     updateThumbnailSize() {
@@ -1172,7 +1253,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             }
         }
         this.removeAll();
-        PopupMenu.PopupMenu.prototype.destroy.call(this);
+        super.destroy();
         this.groupState.disconnect(this.connectId);
         unref(this, RESERVE_KEYS);
     }
