@@ -1,11 +1,11 @@
-#!/usr/bin/python2
+#!/usr/bin/python3
 
 import gi
 gi.require_version("Gtk", "3.0")
-gi.require_version('CDesktopEnums', '3.0')
 
-from gi.repository import Gtk, CDesktopEnums
-from GSettingsWidgets import *
+from gi.repository import Gtk
+from SettingsWidgets import SidePage, GSettingsDependencySwitch, DependencyCheckInstallButton, GSettingsSoundFileChooser
+from xapp.GSettingsWidgets import *
 
 DPI_FACTOR_LARGE         = 1.25
 DPI_FACTOR_NORMAL        = 1.0
@@ -32,7 +32,7 @@ class Module:
 
     def on_module_selected(self):
         if not self.loaded:
-            print "Loading Accessibility module"
+            print("Loading Accessibility module")
 
             self.iface_settings = Gio.Settings(schema_id="org.cinnamon.desktop.interface")
             self.wm_settings = Gio.Settings(schema_id="org.cinnamon.desktop.wm.preferences");
